// This extension was developed by Bart Libert 
//
// Based on code by :
// * Baptiste Saleil http://bsaleil.org/
// * Arnaud Bonatti https://github.com/Obsidien
//
// Licence: GPLv2+
const St = imports.gi.St;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const Main = imports.ui.main;
const Lang = imports.lang;
const Clutter = imports.gi.Clutter;
const Gio = imports.gi.Gio;
const Extension = imports.misc.extensionUtils.getCurrentExtension();
const JsTodo = Extension.imports.jsTodoTxt;
const JsTextFile = Extension.imports.jsTextFile;
const Lib = Extension.imports.lib;
const Shared = Extension.imports.sharedConstants;
const GLib = imports.gi.GLib;

const Gettext = imports.gettext;
const _ = Gettext.domain('todotxt').gettext;

const schema = "org.gnome.shell.extensions.TodoTxt"

let todoTxt;
let metaData;

// TodoTxtManager function
function TodoTxtManager(metadata) {
	let locales = metadata.path + "/locale";
	Gettext.bindtextdomain('todotxt', locales);

	this._init();
}

// Prototype
TodoTxtManager.prototype = {
	schema: null,
	settings: null,
	autoarchive: false,
	__proto__: PanelMenu.Button.prototype,
	connectedSignals: null,
	tasks: null,
	debugMode: false,
	groupBy: null,
	groupUngrouped: false,

	_init: function() {
		PanelMenu.Button.prototype._init.call(this, St.Align.START);

		this.tasks = new Array();
		this.groupedTasksParameter = new Array();
		this.groupedTasksParameter[Shared.NO_GROUPING] = "";
		this.groupedTasksParameter[Shared.GROUP_BY_PROJECTS] = "projects";
		this.groupedTasksParameter[Shared.GROUP_BY_CONTEXTS] = "contexts";

		this.buttonText = new St.Label({
			text: _("[...]")
		});
		this.buttonText.set_style("text-align:center;");
		this.actor.add_actor(this.buttonText);
		this.buttonText.get_parent().add_style_class_name("panelButtonWidth");
		this.schema = schema;

		this._initSettings();

		this._loadSettings();

		this._refresh();
	},

	_createNewTaskEntry: function() {
		this.newTask = new St.Entry({
			name: "newTaskEntry",
			hint_text: _("New task..."),
			track_hover: true,
			can_focus: true
		});

		let tasksMenu = this.menu;
		let buttonText = this.buttonText;

		let entryNewTask = this.newTask.clutter_text;

		let _this = this;

		entryNewTask.connect('key-press-event', function(o, e) {
			let symbol = e.get_key_symbol();
			if (symbol == Clutter.Return) {
				tasksMenu.close();
				buttonText.set_text(_("..."));
				var newTask = new JsTodo.TodoTxtItem(o.get_text());
				_this.addTask(newTask);
			}
		});
	},

	_createTodoItem: function(task) {
		let buttonText = this.buttonText;
		let item = new PopupMenu.PopupMenuItem(_(task.text));
		let projectsLabel = new St.Label({
			text: ""
		});
		let varTodoFile = this.todofile;
		let varDoneFile = this.donefile;
		if (task.projects != null) {
			projectsLabel.set_text(task.projects.join());
		}
		item.addActor(projectsLabel);
		let contextsLabel = new St.Label({
			text: ""
		});
		if (task.contexts != null) {
			contextsLabel.set_text(task.contexts.join());
		}
		item.addActor(contextsLabel);
		let taskClicked = task;
		let _this = this;
		if (!task.complete) {
			let doneButton = new St.Button({
				child: new St.Icon({
					icon_name: 'object-select-symbolic',
					icon_size: 16
				})
			});
			doneButton.connect('clicked', function() {
				buttonText.set_text("[...]");
				_this.completeTask(taskClicked);
			});
			item.addActor(doneButton);
		}
		else {
			let archiveButton = new St.Button({
				child: new St.Icon({
					icon_name: 'document-save-symbolic',
					icon_size: 16
				})
			});
			archiveButton.connect('clicked', function() {
				buttonText.set_text("[...]");
				_this.archiveTask(taskClicked);
			})
			item.actor.add_style_class_name("doneTaskItem");
			item.addActor(archiveButton);
		}
		let deleteButton = new St.Button({
			child: new St.Icon({
				icon_name: 'edit-delete-symbolic',
				icon_size: 16
			})
		});
		deleteButton.connect('clicked', function() {
			buttonText.set_text("[...]");
			_this.removeTask(taskClicked, varTodoFile);
		});
		item.addActor(deleteButton);
		item.connect('activate', function() {
			editTask = new St.Entry({
				name: "editTaskEntry",
				track_hover: true,
				can_focus: true
			});
			editTask.clutter_text.insert_text(task.toString(), 0);
			editTask.clutter_text.set_cursor_position(task.toString().length);

			entryNewTask.set_text('');
			entryNewTask.insert_text(task.toString(), 0);
		});
		return item;
	},

	_initSettings: function() {
		let settings = new Lib.Settings(this.schema);
		this.settings = settings.getSettings();
	},

	_loadSettings: function() {
		this.todofile = this.settings.get_string("todotxt-location").replace("$HOME", GLib.get_home_dir()).replace("~", GLib.get_home_dir());
		this.donefile = this.settings.get_string("donetxt-location").replace("$HOME", GLib.get_home_dir()).replace("~", GLib.get_home_dir());
		this.autoarchive = this.settings.get_boolean("auto-archive");
		this.debugMode = this.settings.get_boolean("debug-mode");
		this.groupBy = this.settings.get_int("group-by");
		this.groupUngrouped = this.settings.get_boolean("group-ungrouped");
	},

	_connectSettingsSignals: function() {
		if (this.connectedSignals == null) {
			this.connectedSignals = new Array();
			this.connectedSignals.push(this.settings.connect("changed::auto-archive", Lang.bind(this, this.onParamChanged)));
			this.connectedSignals.push(this.settings.connect("changed::todotxt-location", Lang.bind(this, this.onParamChanged)));
			this.connectedSignals.push(this.settings.connect("changed::donetxt-location", Lang.bind(this, this.onParamChanged)));
			this.connectedSignals.push(this.settings.connect("changed::debug-mode", Lang.bind(this, this.onParamChanged)));
			this.connectedSignals.push(this.settings.connect("changed::group-by", Lang.bind(this, this.onParamChanged)));
			this.connectedSignals.push(this.settings.connect("changed::group-ungrouped", Lang.bind(this, this.onParamChanged)));
		}
	},

	_refresh: function() {
		let varFile = this.todofile;
		let tasksMenu = this.menu;
		let buttonText = this.buttonText;

		this._createNewTaskEntry();

		// Clear
		tasksMenu.removeAll();

		// Sync
		let todoFile = new JsTextFile.JsTextFile(this.todofile);
		if (!todoFile.exists()) {
			return false;
		}

		this._createTasksFromFile(todoFile.getLines());

		this._createTasksMenu();

		buttonText.set_text("[" + this.tasks.length + "]");
		if (this.tasks.length < 10) buttonText.get_parent().add_style_class_name("panelButtonWidth");
		else buttonText.get_parent().remove_style_class_name("panelButtonWidth");

		// Separator
		this.Separator = new PopupMenu.PopupSeparatorMenuItem();
		tasksMenu.addMenuItem(this.Separator);

		// Bottom section
		let bottomSection = new PopupMenu.PopupMenuSection();

		bottomSection.actor.add_actor(this.newTask);
		bottomSection.actor.add_style_class_name("newTaskSection");
		tasksMenu.addMenuItem(bottomSection);
		tasksMenu.connect('open-state-changed', Lang.bind(this, function(menu, isOpen) {
			if (isOpen) {
				this.newTask.grab_key_focus();
			}
		}));
		return true;
	},

	_createTasksMenu: function() {
		if (this.groupBy == Shared.NO_GROUPING) {
			this.tasks.sort(this.sortByPriority);
			for (let i = 0; i < this.tasks.length; i++) {
				let item = this._createTodoItem(this.tasks[i]);
				this.menu.addMenuItem(item);
			};
			return;
		}
		let groupedTasks = {};
		for (var i = 0; i < this.tasks.length; i++) {
			this._createGroupedTask(this.tasks[i], groupedTasks);
		};
		for (var groupArray in groupedTasks) {
			groupedTasks[groupArray].sort(this.sortByPriority);
		};
		for (var group in groupedTasks) {
			if (group == "__nogroup__") {
				if (this.groupUngrouped) {
					var groupItem = new PopupMenu.PopupSubMenuMenuItem(_("Ungrouped"));
				}
				else {
					var groupItem = this;
				}
			}
			else {
				var groupItem = new PopupMenu.PopupSubMenuMenuItem(group);
			}
			for (let i = 0; i < groupedTasks[group].length; i++) {
				let item = this._createTodoItem(groupedTasks[group][i]);
				groupItem.menu.addMenuItem(item);
			};
			if (groupItem != this) {
				this.menu.addMenuItem(groupItem);
			}
		}
	},

	_createTasksFromFile: function(lines) {
		this.tasks.length = 0;
		for (let i = 0; i < lines.length; i++) {
			if (lines[i] != '' && lines[i] != '\n') {
				let task = new JsTodo.TodoTxtItem(lines[i]);
				this.tasks.push(task);
			}

		};
	},

	_createGroupedTask: function(task, groupedTasks) {
		let groups = task[this.groupedTasksParameter[this.groupBy]];
		if (groups == null) {
			groups = ['__nogroup__'];
		}
		for (var group in groups) {
			if (typeof groupedTasks[groups[group]] == 'undefined') {
				groupedTasks[groups[group]] = [task];
			}
			else {
				groupedTasks[groups[group]].push(task);
			}
		};
	},

	_enable: function() {
		let menuManager = Main.panel._menus || Main.panel.menuManager;
		menuManager.addMenu(this.menu);

		this._connectSettingsSignals();

		// Refresh menu
		let jsTextFile = new JsTextFile.JsTextFile(this.todofile);
		if (jsTextFile.exists()) {
			this.debug("File found: " + this.todofile);
			let fileM = Gio.file_new_for_path(this.todofile);
			this.monitor = fileM.monitor(Gio.FileMonitorFlags.NONE, null);
			this.monitor.connect('changed', Lang.bind(this, this._refresh));
		}
		else {
			this.buttonText.set_text("[X]");
			let errorItem = new PopupMenu.PopupMenuItem(_("No valid todo.txt file specified"));
			this.menu.addMenuItem(errorItem);
		}
	},

	_disable: function() {
		let menuManager = Main.panel._menus || Main.panel.menuManager;
		menuManager.removeMenu(this.menu);
		this.monitor.cancel();
		let varSettings = this.settings;
		if (this.connectedSignals != null) {
			this.connectedSignals.forEach(
			function(signal) {
				if (signal) {
					varSettings.disconnect(signal);
				}
			}

			);
		}
		this.connectedSignals = null;
		this.monitor.cancel();
	},

	onParamChanged: function() {
		this._loadSettings();
		this._refresh();
	},

	removeTask: function(task) {
		let index = this.tasks.indexOf(task);
		if (index == - 1) {
			this.debug("Task not found");
			return false;
		}
		this.tasks.splice(index, 1);
		return this.saveTasksToFile();
	},

	addTask: function(task) {
		this.tasks.push(task);
		return this.saveTasksToFile();
	},

	modifyTask: function(oldTask, newTask, save) {
		let index = this.tasks.indexOf(oldTask);
		if (index == - 1) {
			this.debug("Task not found");
			return false;
		}
		this.tasks[index] = newTask;
		if (save) {
			return this.saveTasksToFile();
		}
		return true;
	},

	completeTask: function(task) {
		let doneTask = new JsTodo.TodoTxtItem(task.toString());
		doneTask.complete = true;
		doneTask.completed = new Date();
		// Modify tasks list, but don't save yet
		if (!this.modifyTask(task, doneTask, false)) {
			return false;
		}
		// If autoarchive is on, archive task and save both files
		if (this.autoarchive) {
			return this.archiveTask(doneTask);
		}
		// If autoarchive is off, only save todo.txt file
		return this.saveTasksToFile();
	},

	archiveTask: function(task) {
		if (!task.complete) {
			this.debug("archiveTask: trying to archive task that is not done")
			return false;
		}
		let jsTextFile = new JsTextFile.JsTextFile(this.donefile);
		if (!jsTextFile.addLine(task.toString())) {
			this.debug("Could not add task to done file");
		}
		if (!jsTextFile.saveFile(true)) {
			this.debug("Could not save done file");
		}
		return this.removeTask(task);
	},

	saveTasksToFile: function() {
		let jsTextFile = new JsTextFile.JsTextFile(this.todofile);
		let textArray = new Array();
		for (let i = 0; i < this.tasks.length; i++) {
			textArray[i] = this.tasks[i].toString();
		}
		jsTextFile.setLines(textArray);
		return jsTextFile.saveFile(true);
	},

	debug: function(text) {
		if (this.debugMode) {
			global.log("[todo.txt] " + text + "\n");
		}
	},

	sortByPriority: function(a, b) {
		if (a.priority == null) {
			if (b.priority == null) {
				return 0;
			}
			// Convention: "null" has smaller priority then everything else
			return 1;
		}
		if (b.priority == null) {
			// Case a==null, b==null already covered
			return - 1;
		}
		return (a.priority.charCodeAt(0) - b.priority.charCodeAt(0));
	}
}

// Init function
function init(metadata) {
	metaData = metadata;
}

function enable() {
	todoTxt = new TodoTxtManager(metaData);
	todoTxt._enable();
	Main.panel.addToStatusArea('todoTxt', todoTxt);
}

function disable() {
	todoTxt._disable();
	todoTxt.destroy();
	todoTxt = null;
}

/* vi: set expandtab tabstop=4 shiftwidth=4: */

